''************************************
''*  Full-Duplex Serial Driver v1.1b *
''*  (C) 2006 Parallax, Inc.         *
''************************************
' 2008.4.12 Eric Ratliff, added some comments
'                         I observe that this object has '8n' parameters only, and fixed quantity of stop and strart bits
' 2008.11.26 Eric Ratliff, added more comments, especially about how to increase rx and tx buffer size, see "SIZE"  in comments
' 2008.11.26 Eric Ratliff, increased receive buffer size to 128 bytes
' 2008.11.27 Eric Ratliff, added way to see buffer heads and tails
' 2008.12.4  Eric Ratliff, added microsecond timeout character get routine, will not support change of clock rate after "start" is called
' 2008.12.6  Eric Ratliff, fixed rxtime to test => 0, not > 0.  Was always waiting 1 ms too long before this fix.

VAR

  long  cog                     'cog flag/id
  long four_us_cl               ' 4 microseconds of time (clocks)

  '9 contiguous longs
  long  rx_head                 ' indicies of where we are in the receive buffer
  long  rx_tail
  long  tx_head                 ' indicies of where we are in the transmit buffer
  long  tx_tail
  long  rx_pin
  long  tx_pin
  long  rxtx_mode
  long  bit_ticks
  long  buffer_ptr

  'transmit and receive buffers
  byte  rx_buffer[128]          ' SIZE may be a power of 2, must follow value of where overflow clip bit is 2 places below
  byte  tx_buffer[16]

  'long dummy0        ' something to assign to


PUB start(rxpin, txpin, mode, baudrate) : okay

'' Start serial driver - starts a cog
'' returns false if no cog available
''
'' mode bit 0 = invert rx
'' mode bit 1 = invert tx
'' mode bit 2 = open-drain/source tx
'' mode bit 3 = ignore tx echo on rx

  stop
  longfill(@rx_head, 0, 4)
  longmove(@rx_pin, @rxpin, 3)  ' copy first three arguments to global variables
  bit_ticks := clkfreq / baudrate
  buffer_ptr := @rx_buffer
  four_us_cl := clkfreq / 250_000 ' pre compute constant for microsecond routine, 4 microseconds as clocks
  okay := cog := cognew(@entry, @rx_head) + 1


PUB GetControlStructure
'' allows external program to monitor buffer fullness
  return @rx_head

  
PUB stop

'' Stop serial driver - frees a cog

  if cog
    cogstop(cog~ - 1)
  longfill(@rx_head, 0, 9)


PUB rxflush

'' Flush receive buffer

  repeat while rxcheck => 0
  
    
PUB rxcheck : rxbyte

'' Check if byte received (never waits)
'' returns -1 if no byte received, $00..$FF if byte

  rxbyte--
  if rx_tail <> rx_head
    rxbyte := rx_buffer[rx_tail]
    rx_tail := (rx_tail + 1) & $7F                      ' must match SIZE of receive buffer, sF for 16 byte buffer, s7F for 128 byte buffer
  'else
    ' fake get of new character
    'rxbyte := rx_buffer[rx_tail]
    'dummy0 := (rx_tail + 1) & $7F                      ' must match SIZE of receive buffer, sF for 16 byte buffer, s7F for 128 byte buffer


PUB rxtime(ms) : rxbyte | t

'' Wait ms milliseconds for a byte to be received
'' returns -1 if no byte received, $00..$FF if byte

  t := cnt
  repeat until (rxbyte := rxcheck) => 0 or (cnt - t) / (clkfreq / 1000) => ms
  

PUB rxtime_us(us) : rxbyte | t, deadline

'' Wait us microseconds for a byte to be received
'' returns -1 if no byte received, $00..$FF if byte
'' does not support change of clock rate after "start" is called or clock frequency < 250 KHz
'' resolution > 4 microseconds
' added partly because I don't like having two divisons in loop, fearing this makes ms wait routine fail at high baud rates such as 115200

  ' fastest check for a character
  if (rxbyte := rxcheck) => 0
    return

  t := cnt  ' record entry time, somewhat delayed
  deadline := (us * four_us_cl) << 2                    
  repeat until (rxbyte := rxcheck) => 0 or (cnt - t) => deadline
  

PUB rx : rxbyte

'' Receive byte (may wait for byte)
'' returns $00..$FF

  repeat while (rxbyte := rxcheck) < 0


PUB tx(txbyte)

'' Send byte (may wait for room in buffer)

  repeat until (tx_tail <> (tx_head + 1) & $F)
  tx_buffer[tx_head] := txbyte
  tx_head := (tx_head + 1) & $F

  if rxtx_mode & %1000
    rx


PUB str(stringptr)

'' Send string                    

  repeat strsize(stringptr)
    tx(byte[stringptr++])
    

PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    tx("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      tx(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      tx("0")
    i /= 10


PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    tx(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    tx((value <-= 1) & 1 + "0")


DAT

'***********************************
'* Assembly language serial driver *
'***********************************

                        org
'
'
' Entry
'
entry                   mov     t1,par                'get structure address
                        add     t1,#4 << 2            'skip past heads and tails

                        rdlong  t2,t1                 'get rx_pin
                        mov     rxmask,#1
                        shl     rxmask,t2

                        add     t1,#4                 'get tx_pin
                        rdlong  t2,t1
                        mov     txmask,#1
                        shl     txmask,t2

                        add     t1,#4                 'get rxtx_mode
                        rdlong  rxtxmode,t1

                        add     t1,#4                 'get bit_ticks
                        rdlong  bitticks,t1

                        add     t1,#4                 'get buffer_ptr
                        rdlong  rxbuff,t1
                        mov     txbuff,rxbuff
                        add     txbuff,#128           'must match SIZE of input buffer

                        test    rxtxmode,#%100  wz    'init tx pin according to mode (z = not tri state?)
                        test    rxtxmode,#%010  wc    '                              (c = invert TX?    )
        if_z_ne_c       or      outa,txmask
        if_z            or      dira,txmask           'make the transmit pin an output now only if we're not doing tri state

                        mov     txcode,#transmit      'initialize ping-pong multitasking
'
'
' Receive
'
receive                 jmpret  rxcode,txcode         'run a chunk of transmit code, then return

                        test    rxtxmode,#%001  wz    'wait for start bit on rx pin
                        test    rxmask,ina      wc    '                              (c = RX pin low?   )
        if_z_eq_c       jmp     #receive

                        mov     rxbits,#9             'ready to receive byte
                        mov     rxcnt,bitticks
                        shr     rxcnt,#1
                        add     rxcnt,cnt                          

:bit                    add     rxcnt,bitticks        'ready next bit period

:wait                   jmpret  rxcode,txcode         'run a chuck of transmit code, then return

                        mov     t1,rxcnt              'check if bit receive period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        test    rxmask,ina      wc    'receive bit on rx pin
                        rcr     rxdata,#1
                        djnz    rxbits,#:bit

                        shr     rxdata,#32-9          'justify and trim received byte
                        and     rxdata,#$FF
                        test    rxtxmode,#%001  wz    'if rx inverted, invert byte
        if_nz           xor     rxdata,#$FF

                        'save received byte and inc head
                        rdlong  t2,par                  ' get the current receive index
                        add     t2,rxbuff               ' add index to buffer base address
                        wrbyte  rxdata,t2               ' (!?) is the address really ready yet? I am surprised there is not inbetween instruction
                        sub     t2,rxbuff               ' get back to the buffer base address
                        add     t2,#1                   ' increment the receeive index
                        and     t2,#$7F                 ' do remanider function for buffer that has a size equal to a power of 2, clips off the overflow bit
                                                        ' depending on SIZE of buffer, #s0F is for 16 byte buffer, $s7F is for 128  bytes
                        wrlong  t2,par                  ' report back the new receive index

                        jmp     #receive                'byte done, receive next byte
'
'
' Transmit
'
transmit                jmpret  txcode,rxcode           'run a chunk of receive code, then return

                        mov     t1,par                  'check for head <> tail
                        add     t1,#2 << 2
                        rdlong  t2,t1
                        add     t1,#1 << 2
                        rdlong  t3,t1
                        cmp     t2,t3           wz
        if_z            jmp     #transmit

                        add     t3,txbuff               'get byte and inc tail
                        rdbyte  txdata,t3                                                        
                        sub     t3,txbuff
                        add     t3,#1
                        and     t3,#$0F                 ' do remanider function for buffer that has a size equal to a power of 2, clips off the overflow bit
                                                        ' depending on SIZE of buffer, #s0F is for 16 byte buffer                                           
                        wrlong  t3,t1
                                                        
                        or      txdata,#$100            'ready byte to transmit
                        shl     txdata,#2
                        or      txdata,#1
                        mov     txbits,#11
                        mov     txcnt,cnt

:bit                    test    rxtxmode,#%100  wz      'output bit on tx pin according to mode
                        test    rxtxmode,#%010  wc
        if_z_and_c      xor     txdata,#1
                        shr     txdata,#1       wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask
                        add     txcnt,bitticks          'ready next cnt

:wait                   jmpret  txcode,rxcode           'run a chunk of receive code, then return

                        mov     t1,txcnt                'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        djnz    txbits,#:bit            'another bit to transmit?

                        jmp     #transmit               'byte done, transmit next byte
'
'
' Uninitialized data
'
t1                      res     1
t2                      res     1
t3                      res     1

rxtxmode                res     1
bitticks                res     1

rxmask                  res     1
rxbuff                  res     1
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1

txmask                  res     1
txbuff                  res     1
txdata                  res     1
txbits                  res     1
txcnt                   res     1
txcode                  res     1